package com.github.huangp.learning.ga.travellingsalemanproblem;

import java.util.Iterator;
import java.util.Map;

import com.google.common.base.Predicate;
import com.google.common.collect.Iterables;
import com.google.common.collect.Maps;
import lombok.extern.slf4j.Slf4j;

/**
 * @author Patrick Huang
 *         <a href="mailto:pahuang@redhat.com">pahuang@redhat.com</a>
 */
@Slf4j
public class TravellingSalesmanGA implements GeneticAlgorithm<CandidateTour, PopulationOfTours> {
    private double mutationRate = 0.015;
    private int selectionSampleSize = 10;

    public TravellingSalesmanGA(double mutationRate, int selectionSampleSize) {
        this.mutationRate = mutationRate;
        this.selectionSampleSize = selectionSampleSize;
    }

    @Override
    public PopulationOfTours evolvePopulation(PopulationOfTours oldPopulation) {
        PopulationOfTours newPop = PopulationOfTours.emptyPopulation(oldPopulation.size());
        // we keep old generation's fittest at first slot
        newPop.setIndividual(0, oldPopulation.getFittest());
//            log.debug("old fittest: {}", oldPopulation.getFittest());
        for (int i = 1; i < oldPopulation.size(); i++) {
            CandidateTour parentOne = selectFittestFromSample(oldPopulation);
//            log.debug("parent one : {}", parentOne);
            CandidateTour parentTwo = selectFittestFromSample(oldPopulation);
//            log.debug("parent two : {}", parentTwo);
            CandidateTour child = crossover(parentOne, parentTwo);
//            log.debug("child      : {}", child);
            CandidateTour mutated = mutate(child);
//            log.debug("mutated    : {}", mutated);
            newPop.setIndividual(i, mutated);
        }
        return newPop;
    }

    @Override
    public CandidateTour crossover(CandidateTour one, CandidateTour other) {
        CandidateTour result = CandidateTour.emptyTour();
        int numOfCities = DestinationCities.manager().numOfCities();
        int rangeOne = (int) (Math.random() * numOfCities);
        int rangeTwo = (int) (Math.random() * numOfCities);
        int rangeStart = Math.min(rangeOne, rangeTwo);
        int rangeEnd = Math.max(rangeOne, rangeTwo);
        // copy over cities from one parent
        final Map<String, City> fromParentOne = Maps.newHashMap();
        for (int i = rangeStart; i < rangeEnd; i++) {
            City city = one.getCity(i);
            result.setCity(i, city);
            fromParentOne.put(city.toString(), city);
        }
        // copy rest from other parent
        Iterator<City> iterator = Iterables.filter(other.getCities(), new Predicate<City>() {
            @Override
            public boolean apply(City input) {
                return !fromParentOne.containsKey(input.toString());
            }
        }).iterator();
        int index = 0;
        while (index < numOfCities) {
            if (result.getCity(index) == null) {
                result.setCity(index, iterator.next());
            }
            index++;
        }
        return result;
    }

    @Override
    public CandidateTour mutate(CandidateTour target) {
        for (int i = 0; i < target.getCities().size(); i++) {
            if (Double.compare(Math.random(), mutationRate) < 0) {
                int indexToSwap = (int) (Math.random() * target.getCities().size());
                City cityOne = target.getCity(i);
                City cityTwo = target.getCity(indexToSwap);
                // swap
                target.setCity(i, cityTwo);
                target.setCity(indexToSwap, cityOne);
            }
        }
        return target;
    }

    @Override
    public CandidateTour selectFittestFromSample(Population<CandidateTour> oldPopulation) {
        PopulationOfTours sample =
                PopulationOfTours.emptyPopulation(selectionSampleSize);
        for (int i = 0; i < selectionSampleSize; i++) {
            int index = (int) (Math.random() * oldPopulation.size());
            CandidateTour candidateTour = oldPopulation.getIndividual(index);
            sample.setIndividual(i, candidateTour);
        }
        return sample.getFittest();
    }
}
